# 视觉小部件


``` python
! [ -e /content ] && pip install -Uqq fastai  # 在Colab上升级fastai
```

``` python
from __future__ import annotations
from fastai.torch_basics import *
from fastai.data.all import *
from fastai.vision.core import *
from fastcore.parallel import *
from ipywidgets import HBox,VBox,widgets,Button,Checkbox,Dropdown,Layout,Box,Output,Label,FileUpload
```

``` python
from nbdev.showdoc import *
```

``` python
_all_ = ['HBox','VBox','widgets','Button','Checkbox','Dropdown','Layout','Box','Output','Label','FileUpload']
```

> 用于图像的ipywidgets

``` python
@patch
def __getitem__(self:Box, i): return self.children[i]
```

``` python
def widget(im, *args, **layout) -> Output:
    "Convert anything that can be `display`ed by `IPython` into a widget"
    o = Output(layout=merge(*args, layout))
    with o: display(im)
    return o
```

``` python
show_doc(widget)
```

------------------------------------------------------------------------

#### widget

>      widget (im, *args, **layout)

Convert anything that can be `display`ed by `IPython` into a widget

``` python
im = Image.open('images/puppy.jpg').to_thumb(256,512)
VBox([widgets.HTML('Puppy'),
      widget(im, max_width="192px")])
```

    VBox(children=(HTML(value='Puppy'), Output(layout=Layout(max_width='192px'))))

``` python
def _update_children(
    change:dict # 一个存储有关已更改小部件信息的字典
):
    "Sets a value to the `layout` attribute on widget initialization and change"
    for o in change['owner'].children:
        if not o.layout.flex: o.layout.flex = '0 0 auto'
```

``` python
def carousel(
    children:tuple|MutableSequence=(), # 用于在轮播图中展示的`Box`对象
    **layout
) -> Box: # An `ipywidget`'s carousel
    "A horizontally scrolling carousel"
    def_layout = dict(overflow='scroll hidden', flex_flow='row', display='flex')
    res = Box([], layout=merge(def_layout, layout))
    res.observe(_update_children, names='children')
    res.children = children
    return res
```

``` python
show_doc(carousel)
```

------------------------------------------------------------------------

#### carousel

>      carousel (children:Union[tuple,list]=(), **layout)

A horizontally scrolling carousel

<table>
<colgroup>
<col style="width: 6%" />
<col style="width: 25%" />
<col style="width: 34%" />
<col style="width: 34%" />
</colgroup>
<thead>
<tr class="header">
<th></th>
<th><strong>Type</strong></th>
<th><strong>Default</strong></th>
<th><strong>Details</strong></th>
</tr>
</thead>
<tbody>
<tr class="odd">
<td>children</td>
<td>tuple | list</td>
<td>()</td>
<td><code>Box</code> objects to display in a carousel</td>
</tr>
<tr class="even">
<td>layout</td>
<td></td>
<td></td>
<td></td>
</tr>
<tr class="odd">
<td><strong>Returns</strong></td>
<td><strong>Box</strong></td>
<td></td>
<td><strong>An <code>ipywidget</code>’s carousel</strong></td>
</tr>
</tbody>
</table>

``` python
ts = [VBox([widget(im, max_width='192px'), Button(description='click')])
      for o in range(3)]

carousel(ts, width='450px')
```

    Box(children=(VBox(children=(Output(layout=Layout(max_width='192px')), Button(description='click', style=Butto…

``` python
def _open_thumb(
    fn:Path|str, # 图像的路径
    h:int, # 缩略图高度
    w:int # 缩略图宽度
) -> Image: # 显示的`PIL`图像
    "Opens an image path and returns the thumbnail of the image"
    return Image.open(fn).to_thumb(h, w).convert('RGBA')
```

``` python
class ImagesCleaner:
    "A widget that displays all images in `fns` along with a `Dropdown`"
    def __init__(self,
        opts:tuple=(), # `Dropdown` 菜单的选项
        height:int=128, # 缩略图高度
        width:int=256, # 缩略图宽度
        max_n:int=30 # 显示的最大图片数量
    ):
        opts = ('<Keep>', '<Delete>')+tuple(opts)
        store_attr('opts,height,width,max_n')
        self.widget = carousel(width='100%')

    def set_fns(self,
        fns:list # 包含每个图像的路径 
    ):
        "Sets a `thumbnail` and a `Dropdown` menu for each `VBox`"
        self.fns = L(fns)[:self.max_n]
        ims = parallel(_open_thumb, self.fns, h=self.height, w=self.width, progress=False,
                       n_workers=min(len(self.fns)//10,defaults.cpus))
        self.widget.children = [VBox([widget(im, height=f'{self.height}px'), Dropdown(
            options=self.opts, layout={'width': 'max-content'})]) for im in ims]

    def _ipython_display_(self): display(self.widget)
    def values(self) -> list:
        "Current values of `Dropdown` for each `VBox`"
        return L(self.widget.children).itemgot(1).attrgot('value')
    def delete(self) -> list:
        "Indices of items to delete"
        return self.values().argwhere(eq('<Delete>'))
    def change(self) -> list:
        "Tuples of the form (index of item to change, new class)"
        idxs = self.values().argwhere(not_(in_(['<Delete>','<Keep>'])))
        return idxs.zipwith(self.values()[idxs])
```

``` python
show_doc(ImagesCleaner)
```

------------------------------------------------------------------------

### ImagesCleaner

>      ImagesCleaner (opts:'tuple'=(), height:'int'=128, width:'int'=256,
>                     max_n:'int'=30)

A widget that displays all images in `fns` along with a `Dropdown`

<table>
<thead>
<tr class="header">
<th></th>
<th><strong>Type</strong></th>
<th><strong>Default</strong></th>
<th><strong>Details</strong></th>
</tr>
</thead>
<tbody>
<tr class="odd">
<td>opts</td>
<td>tuple</td>
<td>()</td>
<td>Options for the <code>Dropdown</code> menu</td>
</tr>
<tr class="even">
<td>height</td>
<td>int</td>
<td>128</td>
<td>Thumbnail Height</td>
</tr>
<tr class="odd">
<td>width</td>
<td>int</td>
<td>256</td>
<td>Thumbnail Width</td>
</tr>
<tr class="even">
<td>max_n</td>
<td>int</td>
<td>30</td>
<td>Max number of images to display</td>
</tr>
</tbody>
</table>

``` python
fns = get_image_files('images')
w = ImagesCleaner(('A','B'))
w.set_fns(fns)
w
```

    Box(children=(VBox(children=(Output(layout=Layout(height='128px')), Dropdown(layout=Layout(width='max-content'…

``` python
w.delete(),w.change()
```

    ((#0) [], (#0) [])

``` python
def _get_iw_info(
    learn,
    ds_idx:int=0 # `learn.dls`中的索引
) -> list:
    "For every image in `dls` `zip` it's `Path`, target and loss"
    dl = learn.dls[ds_idx].new(shuffle=False, drop_last=False)
    probs,targs,preds,losses = learn.get_preds(dl=dl, with_input=False, with_loss=True, with_decoded=True)
    targs = [dl.vocab[t] for t in targs]
    return L([dl.dataset.items,targs,losses]).zip()
```

``` python
@delegates(ImagesCleaner)
class ImageClassifierCleaner(GetAttr):
    "A widget that provides an `ImagesCleaner` for a CNN `Learner`"
    def __init__(self, learn, **kwargs):
        vocab = learn.dls.vocab
        self.default = self.iw = ImagesCleaner(vocab, **kwargs)
        self.dd_cats = Dropdown(options=vocab)
        self.dd_ds   = Dropdown(options=('Train','Valid'))
        self.iwis = _get_iw_info(learn,0),_get_iw_info(learn,1)
        self.dd_ds.observe(self.on_change_ds, 'value')
        self.dd_cats.observe(self.on_change_ds, 'value')
        self.on_change_ds()
        self.widget = VBox([self.dd_cats, self.dd_ds, self.iw.widget])

    def _ipython_display_(self): display(self.widget)
    def on_change_ds(self,change=None):
        "Toggle between training validation set view"
        info = L(o for o in self.iwis[self.dd_ds.index] if o[1]==self.dd_cats.value)
        self.iw.set_fns(info.sorted(2, reverse=True).itemgot(0))
```

``` python
show_doc(ImageClassifierCleaner)
```

------------------------------------------------------------------------

### ImageClassifierCleaner

>      ImageClassifierCleaner (learn, opts:'tuple'=(), height:'int'=128,
>                              width:'int'=256, max_n:'int'=30)

A widget that provides an `ImagesCleaner` for a CNN `Learner`

# 导出 -

``` python
import nbdev; nbdev.nbdev_export()
```

``` python

```
